-- store four configure data files in the flash
-- at power-on, always configure with default file with a17=a16='0', which has no flash access
-- if it fails, the fourth file with flash access will be automatically restarted
-- if this fails again, dclk stays low and other signals go tri-stated.
-- reconfigure could also be started by sending a reboot command. ma(18) must be '1' and ma(15..0) = "A5A5" 
-- address bits ma17 and ma16 decides which file will be used. ma signal are validated by a rising edge on
-- mc_spare signal.  
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;

entity l_ctlcon is
    port (
        init_done: in STD_LOGIC;
        clk: in STD_LOGIC;
        conf_done: in STD_LOGIC;
        porn: in STD_LOGIC;
        mc_spare: in std_logic;
        ma: inout STD_LOGIC_VECTOR(18 downto 0);
        confign: out STD_LOGIC;
        moen: out STD_LOGIC;
        mcen: out STD_LOGIC;
        mwen: out STD_LOGIC;
        dclk: out STD_LOGIC
    );
end l_ctlcon;
architecture my_arch of l_ctlcon is
signal config_n,done,donep,conf_clk: std_logic;
signal rd_flash,en_dclk,reboot,config,fail: std_logic;
signal div8: integer range 0 to 7;
signal dma: unsigned(15 downto 0);
signal id:  std_logic_vector(1 downto 0);
begin
-- flash memory control signals
moen <= '0' when conf_done = '0' and fail = '0' else 'Z';
mcen <= not rd_flash when conf_done = '0' and fail = '0' else 'Z';
mwen <= '1' when conf_done = '0' and fail = '0' else 'Z';
--
ma(18 downto 16) <= '0' & id when conf_done = '0' and fail = '0' else "ZZZ";
magen: for i in dma'range generate
  ma(i) <= dma(i) when conf_done = '0' and fail = '0' else 'Z';
end generate magen;
--
confign <= config_n;
process(mc_spare,config_n)
begin
  if(config_n = '0')then
    reboot <= '0';
  elsif(mc_spare'event and mc_spare = '1')then
    if(ma(15 downto 0) = "1010010110100101")then
      reboot <= ma(18) and done;
    end if;
  end if;
end process;
process(clk,porn)
--process(clk)
begin
  if(porn = '0')then
    id <= "00";
    en_dclk <= '0';
    div8 <= 0;
    dma <= (others => '0');
    rd_flash <= '0';
    config_n <= '0';
    conf_clk <= '0';
    done <= '0';
    fail <= '0';
    config <= '0';
--    confign <= '1';
  elsif(clk'event and clk = '1')then
-- dclk is conf_clk gated by en_dclk
    conf_clk <= not conf_clk;
    dclk <= conf_clk and en_dclk and not (dma(15) and dma(14) and dma(13) and dma(12));
--    dclk <= '0';
    if(conf_clk = '0')then
--      config <= reboot;
      if(reboot = '1' or (done = '0' and dma(15 downto 12) = 15 and id /= "11"))then
        config <= '1';
      end if;
--      config <= reboot or (not done and dma(15) and dma(14) and dma(13) and dma(12) and not(id(1) and id(0)));
--      confign <= config_n or not reboot;
--      confign <= config_n and not (fail and not dma(7));
--      porn <= not (fail and not dma(7));
      if(done = '0' and dma(15 downto 11) = 31 and id = "11")then
        fail <= '1';
      end if;
      if( config_n = '0' )then
        config <= '0';
      elsif( config = '1' )then
        id(1) <= ma(17) or not done;
        id(0) <= ma(16) or not done;
      end if;
      if(done = '1' and config_n = '1')then
        div8 <= 0;
      elsif(fail = '1' and dma(7) = '1')then
        div8 <= 1;
      else
        div8 <= div8 + 1;
      end if;
      if(div8 = 7)then
        en_dclk <= rd_flash and not done;
      end if;
      if(div8 = 0)then
        donep <= init_done and conf_done;
        done <= donep and init_done and conf_done;
        if(dma(6 downto 0) = 127)then
          config_n <= not config; 
          rd_flash <= not done and config_n and not config;
          if(en_dclk = '1')then
            dma(15 downto 7) <= dma(15 downto 7) + 1;
          else
            dma(15 downto 7) <= (others => '0');
          end if;
        end if;
        dma(6 downto 0) <= dma(6 downto 0) + 1;
      end if;
    end if;
  end if;
end process;
end my_arch;
